"""
.. module:: Katna.image_filters.filter
    :platform: Cross Platform
    :synopsis: image_Feature.Filter is an Base class for Image Filters\
         to be used for Katna Image crop Selection functionality all image filter \
        inherit from this base class
"""

import os
import cv2
import numpy as np


class Filter:
    """Base Class for Image filters
    """

    def __init__(self, weight):
        self.weight = weight

    def get_weight(self):
        """gets weight of particular filter

        :return: weight of the filter
        :rtype: float
        """
        return self.weight



"""
.. module:: Katna.feature_list
    :platform: OS X
    :synopsis: This module creates list of features for image module
"""

import os
import cv2
import pathlib
import numpy as np


class FeatureList:
    """Adapter class for features
    """

    def __init__(self):
        self.features = []
        # Getting the absolute path for the directory in which all the feature modules are kept
        features_modules_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "image_features")
        features_modules_list = os.listdir(features_modules_path)

        # Iterating over each feature module, importing it and adding it to features list
        for each_feature in features_modules_list:

            # Checking if the file is a python file and in format: <feature name>_feature.py
            if each_feature.endswith(".py") and len(each_feature.split("_")) == 2:
                splitted_name = each_feature.split(".")

                # Creating the import statement as string and converting it into import command command
                exec(
                    str(
                        "from Katna.image_features."
                        + splitted_name[0]
                        + " import "
                        + "".join([_.title() for _ in splitted_name[0].split("_")])
                    )
                )

                # Creating a dummy object for the imported feature module class and adding it to the features list
                dummy_obj = eval(
                    "".join([_.title() for _ in splitted_name[0].split("_")]) + "()"
                )
                self.features.append(dummy_obj)

    def get_features(self):
        """ Function to get list of all the builtin features

            :return: Returns list of features
            :rtype: python list of features objects
        """
        return self.features



"""
.. module:: Katana.image_features.face_feature
    :platform: OS X
    :synopsis: This module is for for getting Face detection feature Map from input image
"""

import os
import cv2
import math
import numpy as np
import time
import requests
import imutils
import random


from Katna.image_features.feature import Feature
import Katna.config as config


class FaceFeature(Feature):
    """Class for calculating Face detection feature map from input image \n
    Internal Parameters\n
    model_file : Path for downloading model for face detection\n
    prototxt_file : Path for downloading model description prototxt file\n
    confidence : Face detection confidence threshold value
    """

    def __init__(self, weight=1.0):
        """Constructor for this class does following tasks, if not already downloaded\
        , it first downloads face detector model file and prototxt file from public URL\
        ands save it at USER_HOME/.katna directory, or /tmp/.katna directory.\
        After this initializer code initializes internal parameter: \
        min_confidence (for face detection)
        """
        super().__init__(weight)

        self.model_file = config.FaceFeature.model_file
        self.prototxt_file = config.FaceFeature.prototxt_file
        self.cache_subdir = config.FaceFeature.cache_subdir
        self.confidence = config.FaceFeature.confidence

        try:
            self.network_folder_path = os.path.join(os.path.expanduser("~"), ".katna")
            if not os.access(self.network_folder_path, os.W_OK):
                self.network_folder_path = os.path.join("/tmp", ".katna")
            self.datadir = os.path.join(self.network_folder_path, self.cache_subdir)
            if not os.path.exists(self.datadir):
                os.makedirs(self.datadir)

            self.network_model_file_path = os.path.join(self.datadir, self.model_file)

            self.network_prototxt_file_path = os.path.join(
                self.datadir, self.prototxt_file
            )

            if not os.path.exists(self.network_model_file_path):
                self.download_model()

            if not os.path.exists(self.network_prototxt_file_path):
                self.download_proto()

            # print("model file..." + self.network_prototxt_file_path + " Model file " +  self.network_model_file_path)

            self.net = cv2.dnn.readNetFromCaffe(
                self.network_prototxt_file_path, self.network_model_file_path
            )

        except Exception:
            raise FileNotFoundError(
                self.model_file
                + " or "
                + self.prototxt_file
                + " seems to be missing.\
                 Download the file and specify the full path\
                      while initializing FaceFeature class"
            )

    def download_proto(self):
        """Public function for downloading the network weight from the URL link, to be used for
        face detection functionality. 
        Troubleshooting tip: If you get FileNotFound error during face detector initialization,
        initialize the face detector and call this function directly to download the model file from public URL link.
        """
        # create response object
        link = config.FaceFeature.prototxt_download_link
        r = requests.get(link, stream=True)
        # download started
        print("Downloading model file...")
        # if not os.path.isfile(self.network_file_path) or not os.path.exists(self.network_file_path):
        with open(os.path.join(self.datadir, self.prototxt_file), "wb") as f:
            for chunk in r.iter_content(chunk_size=1024 * 1024):
                if chunk:
                    f.write(chunk)
        print("Prototext file downloaded.")

    def download_model(self):
        """Public function for downloading the network weight from the URL link, to be used for
        face detection functionality. 
        Troubleshooting tip: If you get FileNotFound error during face detector initialization,
        initialize the face detector and call this function directly to download the model file from public URL link.
        """
        # create response object
        link = config.FaceFeature.modelfile_download_link
        r = requests.get(link, stream=True)
        # download started
        print("Downloading model file...")
        # if not os.path.isfile(self.network_file_path) or not os.path.exists(self.network_file_path):
        with open(os.path.join(self.datadir, self.model_file), "wb") as f:
            for chunk in r.iter_content(chunk_size=1024 * 1024):
                if chunk:
                    f.write(chunk)
        print("Caffe Model file downloaded.")

    def get_feature_map(self, image):
        """Public function for getting Feature map image from Face detection in input Image

        :param image: input image
        :type image: `numpy array`
        :return: single channel opencv numpy image with feature map from Face detection
        :rtype: numpy array
        """

        # frame = imutils.resize(image, width=400)
        # grab the frame dimensions and convert it to a blob
        frame = image.copy()
        (h, w) = frame.shape[:2]
        blob = cv2.dnn.blobFromImage(
            cv2.resize(frame, (300, 300)), 1.0, (300, 300), (104.0, 177.0, 123.0)
        )
        # pass the blob through the network and obtain the detections and
        # predictions
        self.net.setInput(blob)
        detections = self.net.forward()
        gray = np.zeros((h, w), dtype=np.uint8)

        # loop over the detections
        for i in range(0, detections.shape[2]):
            # extract the confidence (i.e., probability) associated with the
            # prediction
            confidence = detections[0, 0, i, 2]

            # filter out weak detections by ensuring the `confidence` is
            # greater than the minimum confidence
            if confidence < self.confidence:
                continue

            # compute the (x, y)-coordinates of the bounding box for the
            # object
            box = detections[0, 0, i, 3:7] * np.array([w, h, w, h])
            (startX, startY, endX, endY) = box.astype("int")

            # draw the bounding box of the face along with the associated
            # probability
            text = "{:.2f}%".format(confidence * 100)
            y = startY - 10 if startY - 10 > 10 else startY + 10
            cv2.rectangle(gray, (startX, startY), (endX, endY), (120), -1)
            # cv2.putText(gray, text, (startX, y), cv2.FONT_HERSHEY_SIMPLEX, 0.45, (127), 2)

        return gray



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import re
import os
import sys
import setuptools
from distutils.core import Command

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()

if sys.version_info >= (3,10):
   sys.exit("Python version greater than 3.9 not supported because of potentail compatibility issues")

# This will store the models
network_folder_path = os.path.join(os.path.expanduser("~"), ".katna")
if not os.path.exists(network_folder_path):
    os.mkdir(network_folder_path)

# helper functions to make it easier to list dependencies not as a python list, but vertically w/ optional built-in comments to why a certain version of the dependency is listed
def cleanup(x):
    return re.sub(r" *#.*", "", x.strip())  # comments


def to_list(buffer):
    return list(filter(None, map(cleanup, buffer.splitlines())))


# normal dependencies ###
#
# these get resolved and installed via either of these two:
#
#   pip install katna
#   pip install -e .
#
# IMPORTANT: when updating these, please make sure to sync conda/meta.yaml
dep_groups = {
    "core": to_list(
        """
        scipy
        scikit-learn
        scikit-image
        opencv-contrib-python>=3.4.7
        numpy>=1.15
        imutils
        requests
        moviepy >=1.0.1
"""
    )
}

__version__ = None # Explicitly set version.
exec(open('Katna/version.py').read()) # loads __version__

requirements = [y for x in dep_groups.values() for y in x]
setup_requirements = to_list(
    """
    pytest-runner
    setuptools>=36.2
"""
)


# test dependencies ###
test_requirements = to_list(
    """
    pytest
"""
)


setuptools.setup(
    name="katna",
    version=__version__,
    author="keplerlab",
    author_email="keplerwaasi@gmail.com",
    description="Katna is a tool that automates video key/best frames extraction.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/keplerlab/Katna.git",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    setup_requires=setup_requirements,
    tests_require=test_requirements,
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    project_urls={
        "Documentation": "https://katna.readthedocs.io",
        "Source": "https://github.com/keplerlab/Katna",
        "Tracker": "https://github.com/keplerlab/Katna/issues",
    },
    include_package_data=True,
    zip_safe=False,
)



"""
.. module:: Katna.crop_selecter
    :platform: OS X
    :synopsis: This module has functions related to crop selection
"""

from __future__ import print_function

import os
import cv2
import numpy as np
import Katna.config as config


class CropSelector(object):
    """Class for selection of top n crop rectangles from input list of crop rectangles
    ,It also implements filtering functionality, currently following filtering
    method are implemented: Text Detector filter

    :param object: base class inheritance
    :type object: class:`Object`
    """

    def __init__(self):
        self.debug_image = None

    def _sort(self, filtered_crop_list):
        """ Function to sort the filtered crop list

            :param filtered_crop_list: Required number of crop
            :type filtered_crop_list: list of crop rect objects
            :return: Returns list of filtered crop rect
            :rtype: list of crop rect objects
        """
        # reverse = None (Sorts in Ascending order)
        # key is set to sort using second element of
        # sublist lambda has been used

        # print("\n\nUN SORTED LIST\n\n")
        # for f in filtered_crop_list:
        #    print(f, f.score)
        sorted_list = sorted(
            filtered_crop_list, key=lambda x: float(x.score), reverse=True
        )

        # print("\n\nSORTED LIST\n\n")
        # for f in sorted_list:
        #    print(f, f.score)

        return sorted_list

    def select_candidate_crops(
        self, input_image, num_of_crops, input_crop_list, defined_filters, filters=[]
    ):
        """Public Function for CropSelector class: takes list of
        crop rectangles and number of required crops as input and returns list
        of filtered crop retangles as output. Also takes list of filters to be used for filtering out unwanted crop rectangles 
        as optional input.

        :param object: base class inheritance
        :type object: class:`Object`
        :param input_image: input image
        :type input_image: Opencv Numpy Image
        :param input_crop_list: list of input crop in list of crop_rect format
        :type input_crop_list: python list crop_rect data structure
        :param number_of_crop: Required number of crop
        :type: int

        :return: Returns list of filtered crop_rect
        :rtype: python list of crop_rect data structure
        """

        filtered_crop_list = []

        # Has the user applied any filters
        if self.__are_filters_defined(filters) is False:
            filtered_crop_list = self._sort(input_crop_list)
            return self.__topk(filtered_crop_list, num_of_crops)

        # Are user added filters valid
        if self.__are_filters_valid(defined_filters, filters) is False:
            raise TypeError("Added filters do not match predefined filters")

        # Apply the filter to all the crops. This is like scanning the crops with filters
        for defFilter in defined_filters:

            defFilter.set_image(input_image)
            input_crop_list = self.__remove_crops_from_list(defFilter, input_crop_list)

        self.debug_image = input_image

        filtered_crop_list = self._sort(input_crop_list)

        return self.__topk(filtered_crop_list, num_of_crops)

    def __remove_crops_from_list(self, user_filter, input_crop_list):
        """ Private function to remove crops not matching the user added filter.
        
        :param user_filter: filter added by user
        :type user_filter: custom object of type filter

        :param input_crop_list: list of input crop in list of crop_rect format
        :type input_crop_list: python list crop_rect data structure
        
        :return: Returns updated crop list
        :rtype: python list of crop_rect data structure
        """
        # Iterate from last item to first item of python list input_crop_list 
        for i in range(len(input_crop_list) - 1, -1, -1):
            # Remove item from list if filter response is false 
            # please note iterating list from last to first item ensures 
            # you do not remove items are removed in right order
            if user_filter.get_filter_result(input_crop_list[i]) is False:
                input_crop_list.remove(input_crop_list[i])

        return input_crop_list

    def __are_filters_defined(self, filters):
        """Private Function to check if any filters have been added.

        :param filters: User provided filter list. 
        :type defined_filters: list of filter 

        :return: True if filters are added by user
        :rtype: boolean
        """
        # Check filters which are defined actually are present
        if filters != [] or len(filters) > 0:
            return True
        else:
            return False

    def __are_filters_valid(self, defined_filters, filters):
        """Private Function to check if user added filters are valid.

        :param defined_filters: Master List of filters. 
        :type defined_filters: list of filter 
        :param filters: User provided filter list. 
        :type defined_filters: list of filter 

        :return: True if valid filters
        :rtype: boolean
        """

        if defined_filters is not None or len(defined_filters) > 0:
            for defFilter in defined_filters:
                # Check if user asked filters **filters** are in library implemented filters **defFilter**
                if not (type(defFilter).__name__ in filters):
                    return False
        # if reached here then added filters are valid
        return True

    def __topk(self, crops_list, k):
        """Private Function to return top k items from the crops list.

        :param crops_list: crops left after applying all the filters. 
        :type crops_list: list of crops 
        :param K: number of crops to return. 
        :type K: int 

        :return: crops
        :rtype: list of crops
        """
        return crops_list[:k]



"""
.. module:: Katna.video
    :platform: OS X
    :synopsis: This module has functions related to key frame extraction 
"""
import os.path
import os
import sys
import numpy as np
import cv2
import errno
from Katna.decorators import VideoDecorators
from Katna.decorators import FileDecorators

from Katna.frame_extractor import FrameExtractor
from Katna.image_selector import ImageSelector
import Katna.config as config

from multiprocessing import Pool, Process, cpu_count

from moviepy.editor import VideoFileClip

from moviepy.tools import subprocess_call
import moviepy.config as moviepyConfig
class Video(object):
    """Class for all video frames operations

    :param object: base class inheritance
    :type object: class:`Object`
    """

    def __init__(self):
        # If the duration of the clipped video is less than **min_video_duration**
        # then, the clip will be added with the previous clipped
        self.min_video_duration = config.Video.min_video_duration
        # Creating the multiprocessing pool
        self.n_processes = cpu_count() // 2 - 1
        if self.n_processes < 1:
            self.n_processes = None
        self.pool_extractor = Pool(processes=self.n_processes)
        self.pool_selector = Pool(processes=self.n_processes)

        # Folder to save the videos after clipping
        self.temp_folder = os.path.abspath(os.path.join("clipped"))
        if not os.path.isdir(self.temp_folder):
            os.mkdir(self.temp_folder)

    def _remove_clips(self, video_clips):
        """Remove video clips from the temp directory
        """
        for clip in video_clips:
            os.remove(clip)
            # print(clip, " removed!")

    @FileDecorators.validate_dir_path
    def extract_frames_as_images_from_dir(self, no_of_frames, dir_path):
        """Returns a dict of best key images/frames from the videos in the directory.

        :param no_of_frames: Number of key frames to be extracted
        :type no_of_frames: int, required
        :param dir_path: Directory location with videos
        :type dir_path: str, required
        :return: Dictionary with key as filepath and numpy.2darray Image objects
        :rtype: dict
        """

        all_videos_top_frames = {}
        for path, subdirs, files in os.walk(dir_path):
            for filename in files:
                filepath = os.path.join(path, filename)
                if filename.lower().endswith(".mp4") or filename.lower().endswith(".mov"):
                    video_file_path = os.path.join(path, filename)
                    video_top_frames = self.extract_frames_as_images(no_of_frames, video_file_path)
                    all_videos_top_frames[filepath] = video_top_frames

        return all_videos_top_frames

    @FileDecorators.validate_file_path
    def extract_frames_as_images(self, no_of_frames, file_path):
        """Returns a list of best key images/frames from the video.
        
        :param no_of_frames: Number of key frames to be extracted
        :type no_of_frames: int, required
        :param file_path: video file location
        :type file_path: str, required
        :return: List of numpy.2darray Image objects
        :rtype: list
        """

        # Split the videos into chunks. Each split(video) will be stored in a temp location
        videos = self._split(file_path)

        frame_extractor = FrameExtractor()

        # Passing all the clipped videos for  the frame extraction using map function of the
        # multiprocessing pool
        extracted_candidate_frames = self.pool_extractor.map(
            frame_extractor.extract_candidate_frames, videos
        )
        # Converting the nested list of extracted frames into 1D list
        extracted_candidate_frames = [
            frame for frames in extracted_candidate_frames for frame in frames
        ]

        self._remove_clips(videos)

        image_selecter = ImageSelector(self.pool_selector)
        top_frames = image_selecter.select_best_frames(
            extracted_candidate_frames, no_of_frames
        )

        return top_frames

    @FileDecorators.validate_file_path
    def save_frame_to_disk(self, frame, file_path, file_name, file_ext):
        """saves an in-memory numpy image array on drive.
        
        :param frame: In-memory image. This would have been generated by extract_frames_as_images method
        :type frame: numpy.ndarray, required
        :param file_name: name of the image.
        :type file_name: str, required
        :param file_path: Folder location where files needs to be saved
        :type file_path: str, required
        :param file_ext: File extension indicating the file type for example - '.jpg'
        :type file_ext: str, required         
        :return: None
        """

        file_full_path = os.path.join(file_path, file_name + file_ext)
        cv2.imwrite(file_full_path, frame)

    @FileDecorators.validate_file_path
    def _split(self, file_path):
        """Function to split the videos and persist the chunks

        :param file_path: path of video file
        :type file_path: str, required
        :return: List of path of splitted video clips
        :rtype: list
        """
        clipped_files = []
        duration = VideoFileClip(file_path, audio=False).duration
        # setting the start point to zero
        # Setting the breaking point for the clip to be 25 or if video is big
        # then relative to core available in the machine
        clip_start, break_point = (
            0,
            duration // cpu_count() if duration // cpu_count() > 15 else 25,
        )

        # Loop over the video duration to get the clip stating point and end point to split the video
        while clip_start < duration:

            clip_end = clip_start + break_point
            
            # Setting the end position of the particular clip equals to the end time of original clip,
            # if end position or end position added with the **min_video_duration** is greater than 
            # the end time of original video
            if clip_end > duration or (clip_end + self.min_video_duration) > duration:
                clip_end = duration

            clipped_files.append(
                self.__write_videofile(file_path, clip_start, clip_end)
            )

            clip_start = clip_end
        return clipped_files

    def __write_videofile(self, video_file_path, start, end):
        """Function to clip the video for given start and end points and save the video

        :param video_file_path: path of video file
        :type video_file_path: str, required
        :param start: start time for clipping
        :type start: float, required
        :param end: end time for clipping
        :type end: float, required
        :return: path of splitted video clip
        :rtype: str
        """

        name = os.path.split(video_file_path)[1]

        # creating a unique name for the clip video 
        # Naming Format: <video name>_<start position>_<end position>.mp4
        _clipped_file_path = os.path.join(
            self.temp_folder,
            "{0}_{1}_{2}.mp4".format(
                name.split(".")[0], int(1000 * start), int(1000 * end)
            ),
        )

        # Calling the ffmpeg to extract the clip with start & end position and
        # save it to the target location(temp dir)
        self._ffmpeg_extract_subclip(
            video_file_path, start, end, targetname=_clipped_file_path
        )
        return _clipped_file_path

    def _ffmpeg_extract_subclip(self, filename, t1, t2, targetname=None):
        """ makes a new video file playing video file ``filename`` between
            the times ``t1`` and ``t2``. 
        :param filename: path of video file
        :type filename: str, required
        :param t1: time from where video to clip
        :type t1: float, required
        :param t2: time to which video to clip
        :type t2: float, required
        :param targetname: path where clipped file to be stored
        :type targetname: str, optional
        :return: None
        """
        name, ext = os.path.splitext(filename)

        if not targetname:
            T1, T2 = [int(1000 * t) for t in [t1, t2]]
            targetname = name + "{0}SUB{1}_{2}.{3}".format(name, T1, T2, ext)

        cmd = [
            moviepyConfig.FFMPEG_BINARY,
            "-y",
            "-i",
            filename,
            "-ss",
            "%0.2f" % t1,
            "-t",
            "%0.2f" % (t2 - t1),
            "-vcodec",
            "copy",
            "-acodec",
            "copy",
            targetname,
        ]

        subprocess_call(cmd, logger=None, errorprint=True)
